/*
 * extractdata.c
 * David Moore <dcm@csail.mit.edu>
 *
 * Reads raw output from a cricket and decodes all events that occured
 * during the run such as range measurements, localization, etc.
 * Input to a simulation is produced, and optionally other useful
 * data.
 *
 *
 * Copyright (C) 2004  Massachusetts Institute of Technology
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "parse.h"

#define MAX_HANDLES 256

struct node_info {
    unsigned char   id[4];
};

int handles[MAX_HANDLES];
int num_nodes = 0;
struct node_info * nodes = NULL;
struct timeval lasttime;
int loc_count = 0;
int do_extract = 0;
char outdir[256];
int do_truth = 0;
FILE * simout = NULL;

float ** dist_truth;
float ** dist;

#define square(a) ((a)*(a))

int match_id(unsigned char * id1, unsigned char * id2)
{
    int i;
    for (i=0; i<4; i++) {
        if (id1[i] != id2[i])
            return 0;
    }
    return 1;
}

void copy_id(unsigned char * dst, unsigned char * src)
{
    int i;
    for (i=0; i<4; i++) {
        dst[i] = src[i];
    }
}

int range(event_range * e, struct timeval * tv)
{
    if (simout) {
        if (lasttime.tv_sec != tv->tv_sec ||
                lasttime.tv_usec != tv->tv_usec) {
            fprintf(simout, "time %04ld.%03ld\n", tv->tv_sec,
                    tv->tv_usec / 1000);
        }
        fprintf(simout, "%d %d %d\n",
                handles[e->node_recv], handles[e->node_send], e->dist);
    }
    if (do_truth && e->summary) {
        dist[handles[e->node_recv]][handles[e->node_send]] = e->dist / 30.0;
    }
    lasttime.tv_sec = tv->tv_sec;
    lasttime.tv_usec = tv->tv_usec;
    return 0;
}

void clear_dist()
{
    int i, j;
    
    for (i=0; i<num_nodes; i++) {
        for (j=0; j<num_nodes; j++) {
            dist[i][j] = 0.0;
        }
    }
}

void compute_dist_error()
{
    int num_dist = 0;
    int i,j;
    double sum = 0;

    for (i=0; i<num_nodes; i++) {
        for (j=0; j<num_nodes; j++) {
            if (dist[i][j] != 0.0 && dist[i][j] < dist_truth[i][j]*1.3 &&
                    dist[i][j] > dist_truth[i][j]*0.7) {
                sum += square(dist[i][j]-dist_truth[i][j]);
                num_dist++;
            }
        }
    }
    fprintf(stderr, "   num_dist = %d  sum = %f  err^2 = %f\n", num_dist,
            sum, sum / (num_dist-1));
}

int localize(event_localize * e, struct timeval * tv)
{
    int i, j;
    char filename[256];
    FILE * out;

    fprintf(stderr, "%04d: Localization of %d nodes, %d quads, %d optimized\n",
            loc_count, e->num_pos, e->num_quad, e->num_opt);
    if (do_extract) {
        snprintf(filename, sizeof(filename), "%s/loc%04d", outdir, loc_count);
        out = fopen(filename, "w");
        if (!out)
            return 0;
        fprintf(out, "0.0   0.0     1\n");
        for (i=1; i<num_nodes; i++) {
            for (j=0; j < e->num_pos; j++) {
                if (handles[e->pos[j].node] == i)
                    break;
            }
            if (j == e->num_pos)
                fprintf(out, "0.0   0.0     0\n");
            else
                fprintf(out, "%f %f     1\n", e->pos[j].x, e->pos[j].y);
        }
        fclose(out);
    }

    if (do_truth) {
        compute_dist_error();
        clear_dist();
    }
    loc_count++;
    return 0;
}

void write_index()
{
    char filename[256];
    FILE * out;
    int i;

    snprintf(filename, sizeof(filename), "%s/index", outdir);
    out = fopen(filename, "w");
    if (!out)
        return;

    for (i=0; i<num_nodes; i++) {
        unsigned char * id = nodes[i].id;
        
        fprintf(out, "%02x:%02x:%02x:%02x\n", id[0], id[1], id[2], id[3]);
    }

    fclose(out);
}

int read_truth(char * filename)
{
    FILE * infile;
    float xs[num_nodes], ys[num_nodes];
    int set[num_nodes];
    int id[4];
    unsigned char idt[4];
    int i, j;
    char buf[256];
    float x, y;

    for (i=0; i<num_nodes; i++)
        set[i] = 0;

    infile = fopen(filename, "r");
    if (!infile)
        return -1;

    while (fgets(buf, sizeof(buf), infile)) {
        if (sscanf(buf, "%x:%x:%x:%x %f %f", id, id+1, id+2, id+3,
                    &x, &y) != 6)
            continue;

        idt[0] = id[0];
        idt[1] = id[1];
        idt[2] = id[2];
        idt[3] = id[3];
        for (i=0; i<num_nodes; i++) {
            if (match_id(idt, nodes[i].id)) {
                set[i] = 1;
                xs[i] = x;
                ys[i] = y;
                break;
            }
        }
    }
    fclose(infile);

    for (i=0; i<num_nodes; i++) {
        if (!set[i]) {
            fprintf(stderr, "Error: ground truth is incomplete\n");
            return -1;
        }
    }

    for (i=0; i<num_nodes; i++) {
        for (j=i+1; j<num_nodes; j++) {
            dist_truth[i][j] = dist_truth[j][i] =
                sqrt(square(xs[i]-xs[j])+square(ys[i]-ys[j]));
        }
    }
    return 0;
}

int newnode(event_newnode * e, struct timeval * tv)
{
    int i;
    if (e->self && !match_id(nodes[0].id, e->id)) {
        copy_id(nodes[0].id, e->id);
        handles[e->node] = 0;
        fprintf(stderr, "node %02x:%02x:%02x:%02x with handle %d assigned to slot %d\n",
                e->id[0], e->id[1], e->id[2], e->id[3], e->node, 0);
        if (simout)
            fprintf(simout, "node %d ID %02x:%02x:%02x:%02x\n",
                    0, e->id[0], e->id[1], e->id[2], e->id[3]);
        return 0;
    }

    for (i=0; i<num_nodes; i++) {
        if (match_id(e->id, nodes[i].id))
            break;
    }
    if (i == num_nodes) {
        nodes = realloc(nodes, (num_nodes+1)*sizeof(struct node_info));
        copy_id(nodes[i].id, e->id);
        fprintf(stderr, "node %02x:%02x:%02x:%02x with handle %d assigned to slot %d\n",
                e->id[0], e->id[1], e->id[2], e->id[3], e->node, i);
        if (simout)
            fprintf(simout, "node %d ID %02x:%02x:%02x:%02x\n",
                    i, e->id[0], e->id[1], e->id[2], e->id[3]);
        num_nodes++;
    }
    handles[e->node] = i;
    return 0;
}

int main(int argc, char ** argv)
{
    char c;
    FILE * input = NULL;
    parse_info * parse;
    int i;
    int verbose = 0;
    char truthfile[256];
    struct stat sbuf;
    strcpy(outdir, "");

    while ((c = getopt(argc, argv, "i:l:o:vt:")) >= 0) {
        switch (c) {
            case 'i':
                input = fopen(optarg, "r");
                if (!input) {
                    perror("fopen");
                    exit(1);
                }
                break;
            case 'l':
                do_extract = 1;
                strcpy(outdir, optarg);
                if (stat(outdir, &sbuf)) {
                    if (mkdir(outdir, 0777)) {
                        perror("mkdir");
                        exit(1);
                    }
                }
                else if (!S_ISDIR(sbuf.st_mode)) {
                    fprintf(stderr, "\"%s\" is not a directory\n", outdir);
                    exit(1);
                }
                break;
            case 'o':
                simout = fopen(optarg, "w");
                if (!simout) {
                    perror("fopen");
                    exit(1);
                }
                break;
            case 'v':
                verbose = 1;
                break;
            case 't':
                strcpy(truthfile, optarg);
                do_truth = 1;
                break;
            default:
                exit(1);
        }
    }

    if (!input) {
        fprintf(stderr, "Must specify cricket data file to be read\n");
        exit(1);
    }

    for (i=0; i<MAX_HANDLES; i++)
        handles[i] = -1;
    nodes = (struct node_info *) malloc(sizeof(struct node_info));
    if (!nodes)
        exit(1);
    bzero(nodes, sizeof(struct node_info));
    num_nodes = 1;
    lasttime.tv_sec = -1;
    lasttime.tv_usec = -1;

    parse = parser_init_file(input, 0, 0, verbose?stdout:NULL);
    if (!parse)
        exit(1);

    /* Do one pass of input just to get the full list of nodes */
    parse->cb_newnode = newnode;
    parser_mainloop(parse);

    if (do_truth) {
        dist_truth = (float **) malloc(num_nodes*sizeof(float *));
        dist_truth[0] = (float *) malloc(num_nodes*num_nodes*sizeof(float));
        for (i=1; i<num_nodes; i++)
            dist_truth[i] = dist_truth[0] + i*num_nodes;
        dist = (float **) malloc(num_nodes*sizeof(float *));
        dist[0] = (float *) malloc(num_nodes*num_nodes*sizeof(float));
        for (i=1; i<num_nodes; i++)
            dist[i] = dist[0] + i*num_nodes;

        if (read_truth(truthfile) == -1)
            exit(1);
        clear_dist();
    }

    rewind(input);
    if (do_extract) {
        write_index();
    }
    parse->cb_localize = localize;
    parse->cb_range = range;
    parser_mainloop(parse);
    parser_free(parse);

    fprintf(stderr, "\nUnique nodes: %d, Runtime: %ld.%01lds\n", num_nodes,
            lasttime.tv_sec, lasttime.tv_usec/100000);

    fclose(input);
    if (simout)
        fclose(simout);

    return 0;
}
