function process_results(name, sigma_d)

net = load([name '.mat']);
N = size(net.xy,1);

fprintf('Reading localization results...\n');
cluster_xy = cell(N,1);
for i=1:N,
    s = sprintf('%s/node%04d.out', name, i);
    cluster_xy{i} = load(s);
end

fprintf('Computing inter-node transformations...\n');
cluster_tx = cell(N);
for i=1:N,
    if isempty(cluster_xy{i})
        continue;
    end

    % For transformation purposes, ignore nodes with weak localizations
    %m = find(cluster_xy{i}(:,4) == 0);
    %cxy = cluster_xy{i}(m,:);
    cxy = cluster_xy{i};

    % Consider each localized neighbor j
    for m=1:size(cxy,1),
        j = cxy(m,1);
        if ~isempty(cluster_tx{i,j})
            % already found transformation
            continue;
        end

        % Find the list of localizations performed by the neighbor j
        %n = find(cluster_xy{j}(:,4) == 0);
        %cxy2 = cluster_xy{j}(n,:);
        cxy2 = cluster_xy{j};

        if isempty(cxy2)
            % opposite localization is not present
            cluster_tx{i,j} = [0];
            cluster_tx{j,i} = [0];
            continue;
        end
        % Find the localization of node i by node j.
        n = find(cxy2(:,1) == i);
        if isempty(n)
            % opposite localization is not present
            cluster_tx{i,j} = [0];
            cluster_tx{j,i} = [0];
            continue;
        end

        % Remember localization of node j by i and node i by j.
        axyb = cxy2(n,2:3);
        bxya = cxy(m,2:3);
        % Consider all other localizations
        cxy1 = cxy([1:m-1 m+1:end],:);
        cxy2 = cxy2([1:n-1 n+1:end],:);
        % Find the neighbors in common of the two localizations (support)
        [support,ia,ib] = intersect(cxy1(:,1), cxy2(:,1));
        if isempty(support)
            cluster_tx{i,j} = [0];
            cluster_tx{j,i} = [0];
            %fprintf('%d %d no support\n', i, j);
            continue;
        end

        % Make sure at least one neighbor in common forms a robust triangle
        robust_sup = [];
        for o = 1:length(support),
            s = support(o);
            [r,v]=robust_triangle(net.dist(i,j),net.dist(i,s),...
                net.dist(j,s),sigma_d);
            if r
                robust_sup = o;
                break;
            end
        end
        if isempty(robust_sup)
            cluster_tx{i,j} = [0];
            cluster_tx{j,i} = [0];
            %fprintf('%d %d no robust support\n', i, j);
            continue;
        end

        % The robust neighbor in common is used to determine if a flip
        % is necessary.
        sxya = cxy1(ia(robust_sup),2:3);
        sxyb = cxy2(ib(robust_sup),2:3);
        cross1 = sxya(1)*bxya(2) - sxya(2)*bxya(1);
        cross2 = (sxyb(1)-axyb(1))*(-axyb(2)) - (sxyb(2)-axyb(2))*(-axyb(1));
        if (cross1 * cross2 < 0)
            flip = -1;
        else
            flip = 1;
        end

        % Estimate the translation and rotation
        tx = [bxya atan2(bxya(2),bxya(1))-atan2(-axyb(2),-flip*axyb(1))];
        opts = optimset('Display', 'off');
        % Compute the best translation and rotation using all support
        tx = lsqnonlin(@lsqfun, tx, [], [], opts, ...
            [axyb' [0; 0] cxy2(ib,2:3)'], [[0; 0] bxya' cxy1(ia,2:3)'], flip);

        % Store the results
        cluster_tx{i,j} = tx_to_matrix(tx,flip);
        cluster_tx{j,i} = inv(cluster_tx{i,j});
    end
end

fprintf('Generating routing table...\n');
routing = zeros(N);
weak_routing = zeros(N);
for i = 1:N,
    list = [i];
    routing(i,i) = i;
    while ~isempty(list)
        curr = list(1);
        list = list(2:end);
        if isempty(cluster_xy{curr})
            continue;
        end
        for next = cluster_xy{curr}(:,1)',
            if routing(i,next)
                continue;
            end
            if isempty(cluster_tx{curr,next}) || isequal(cluster_tx{curr,next},[0])
                if ~weak_routing(i,next)
                    weak_routing(i,next) = curr;
                end
                continue;
            end
            routing(i,next) = curr;
            list = [list next];
        end
    end
end

save([name '_tx'], 'cluster_xy', 'cluster_tx', 'routing', 'weak_routing');

%i = 3;
%figure;
%plot(net.xy([1:i-1 i+1:end],1), net.xy([1:i-1 i+1:end],2), '+');
%hold on;
%for j = 1:N,
%    k = routing(i,j);
%    if k ~= 0
%        plot([net.xy(j,1) net.xy(k,1)], [net.xy(j,2) net.xy(k,2)], '-');
%    end
%    k = weak_routing(i,j);
%    if k ~= 0
%        plot([net.xy(j,1) net.xy(k,1)], [net.xy(j,2) net.xy(k,2)], 'g-');
%    end
%end
%plot(net.xy(i,1), net.xy(i,2), 'ro');

function [r,v] = robust_triangle(a, b, c, sigma_d)
e = sort([a b c]);
costh = (e(2)*e(2)+e(3)*e(3)-e(1)*e(1))/(2*e(2)*e(3));
v = e(1)*(1-costh*costh);
if v < 3.0*sigma_d
    r = 0;
else
    r = 1;
end


function e = lsqfun(tx, a, b, flip);
a(1,:) = flip*a(1,:);
a = transform(tx, a);
e = a-b;
e = [e(1,:) e(2,:)]';

function b = transform(tx, a);
b = [cos(tx(3)) -sin(tx(3)); sin(tx(3)) cos(tx(3))] * a;
b = [b(1,:)+tx(1); b(2,:)+tx(2)];

function t = tx_to_matrix(tx, flip);
t = [flip*cos(tx(3)) -sin(tx(3)) tx(1); flip*sin(tx(3)) cos(tx(3)) tx(2); 0 0 1];
